import weka.classifiers.*;
import weka.classifiers.bayes.*;
import weka.classifiers.functions.*;
import weka.classifiers.functions.supportVector.*;
import weka.classifiers.lazy.*;
import weka.classifiers.meta.*;
import weka.classifiers.rules.*;
import weka.classifiers.trees.*;
import weka.core.*;
import weka.experiment.*;
import weka.filters.*;
import weka.filters.unsupervised.attribute.*;
import weka.core.converters.*;
import java.io.*;
import java.util.*;
import weka.filters.supervised.instance.*;

public class ProduceModels 
{
    public double TR;  // Percentage of the training set
    public double PR;  // Percentage of the pruning set
    public double TE;  // Percentage of the test set
    public int M;      // Number of models

    public Instances Train;
    public Instances Pruning;
    public Instances Test;
    public Instances Train80;

    public ProduceModels()
    {
    }

    private Instances LoadArff(String FileStem)
    {
	try {
	    ArffLoader aLoader = new ArffLoader();
	    FileInputStream aFile = new FileInputStream(FileStem);
	    aLoader.setSource(aFile);
	    Instances Data = new Instances(aLoader.getDataSet());
	    aFile.close();
	    return Data;
	}
	catch (Exception Ex) {
	    System.out.println(Ex);
	    System.out.println(FileStem);
	    return null;
	}
    }

    // Save an Instances object in arff format
    private void SaveData(String FileStem,Instances data)
    {
	try
	{
	    File f = new File(FileStem);
	    FileWriter fw = new FileWriter(f);
	    fw.write(data.toString());
	    fw.close();
	}
	catch (Exception Ex){
	    System.out.println(Ex);
	    System.out.println(FileStem);
	}
    }


    public void CreateData(String dataset,String SourcePath, String TargetPath,Random rnd)
    {
	try{

	    Instances Data = LoadArff(SourcePath+dataset+".arff");
	    Data.setClassIndex(Data.numAttributes()-1);
	    Data.randomize(rnd);

	    int num_ins = Data.numInstances();

	    int tr = (int)(TR*num_ins);
	    int pr = (int)(PR*num_ins);
	    int te = (int)(TE*num_ins);

	    //	System.out.println("TR:"+tr+" PR:" + pr +" TE:"+te);

	    Train   = new Instances(Data, 1,tr);
	    Pruning = new Instances(Data, tr, pr);
	    Test    = new Instances(Data,tr+pr,te);
	    Train80 = new Instances(Data,1,tr+pr);

	    SaveData(TargetPath+dataset+"_train"+".arff",Train);
	    SaveData(TargetPath+dataset+"_pruning"+".arff",Pruning);
	    SaveData(TargetPath+dataset+"_test"+".arff",Test);
	    SaveData(TargetPath+dataset+"_train80"+".arff",Train80);


	    CreateModels(Train,TargetPath+"Predictions/",false);
	    CreateModels(Train80,TargetPath+"Predictions/",true);
	}catch(Exception ex){System.out.println(ex);ex.printStackTrace();}
    }

    // Create M models on a dataset and save them
    // If type is true then the train data consist of both train and prune data
    public void CreateModels(Instances data,String path,boolean type)
    {
	Random rand  = new Random(System.currentTimeMillis());

	int id=0;
            // ANN  7x4=28 models
            int[] nodes = {1, 2, 4, 8, 16, 32, 64};
            double[] momentum = {0.0, 0.2, 0.5, 0.9};
            for (int nodesCount=0; nodesCount<nodes.length; nodesCount++)
	    {
		for (int momCount=0; momCount<momentum.length; momCount++)
		{
		    MultilayerPerceptron nn = new MultilayerPerceptron();
		    nn.setHiddenLayers(Integer.toString(nodes[nodesCount]));
		    nn.setMomentum(momentum[momCount]);
		    try{
		    nn.buildClassifier(data);
		    }
		    catch(Exception ex){ex.printStackTrace();System.out.println(ex);}

		    if(type)
		    {
			ResultsDataset predTrainData = new ResultsDataset();
			predTrainData.CreateDataset(nn,Train80,path+"train"+id+".pred");

			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(nn,Test,path+"test"+id+".pred");
		    }

		    if(!type)
		    {
			ResultsDataset predPruneData = new ResultsDataset();
			predPruneData.CreateDataset(nn,Pruning,path+"prune"+id+".pred");

			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(nn,Test,path+"testB"+id+".pred");
		    }
                    id++;
		}
	    }
	    
            // kNN 3x22=66 models
            int numK = 22;
            SelectedTag[] stag = new SelectedTag[3];
            stag[0] = new SelectedTag(IBk.WEIGHT_NONE,IBk.TAGS_WEIGHTING);
            stag[1] = new SelectedTag(IBk.WEIGHT_INVERSE,IBk.TAGS_WEIGHTING);
            stag[2] = new SelectedTag(IBk.WEIGHT_SIMILARITY,IBk.TAGS_WEIGHTING);
	    for(int dwCount=0; dwCount<3; dwCount++)
	    {
    	        int k=1;
		for(int kCount=0; kCount<22; kCount++)
		{
		    IBk ibkModel = new IBk(k);
		    ibkModel.setDistanceWeighting(stag[dwCount]);
		    try{
			ibkModel.buildClassifier(data);
		    }
		    catch(Exception ex){ex.printStackTrace();System.out.println(ex);}
		    if(type)
		    {
			ResultsDataset predTrainData = new ResultsDataset();
			predTrainData.CreateDataset(ibkModel,Train80,path+"train"+id+".pred");

			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(ibkModel,Test,path+"test"+id+".pred");
		    }
		    else
		    {
			ResultsDataset predPruneData = new ResultsDataset();
			predPruneData.CreateDataset(ibkModel,Pruning,path+"prune"+id+".pred");

			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(ibkModel,Test,path+"testB"+id+".pred");
		    }
		    if (k == 1)
			k += (data.numInstances()/numK)-1;
		    else
			k += (data.numInstances()/numK);
		    id++;
		}
	    }       
            
            // SVM = 10x10=100 models
            double[] c = {10E-7, 10E-6, 10E-5, 10E-4, 0.001, 0.01, 0.1, 1, 10, 100};
            CachedKernel[] ker =  new CachedKernel[10];
            ker[0] = new PolyKernel();
            ((PolyKernel) ker[0]).setExponent(2.0);
            ker[1] = new PolyKernel();
            ((PolyKernel) ker[1]).setExponent(3.0);
            ker[2] = new RBFKernel();
            ((RBFKernel) ker[2]).setGamma(0.001);
            ker[3] = new RBFKernel();
            ((RBFKernel) ker[3]).setGamma(0.005);
            ker[4] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[4]).setGamma(0.01);
            ker[5] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[5]).setGamma(0.05);
            ker[6] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[6]).setGamma(0.1);
            ker[7] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[7]).setGamma(0.5);
            ker[8] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[8]).setGamma(1);
            ker[9] = new weka.classifiers.functions.supportVector.RBFKernel();
            ((RBFKernel) ker[9]).setGamma(2);

	    for(int cCount=0; cCount<c.length; cCount++)
	    {
                SMO svm = new SMO();
                svm.setC(c[cCount]);
                for (int kCount=0; kCount<ker.length; kCount++)
		{
		    svm.setKernel(ker[kCount]);

		    try{
		    svm.buildClassifier(data);
		    }
		    catch(Exception ex){ex.printStackTrace();System.out.println(ex);}
		    if(type)
		    {
			ResultsDataset predTrainData = new ResultsDataset();
			predTrainData.CreateDataset(svm,Train80,path+"train"+id+".pred");
			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(svm,Test,path+"test"+id+".pred");
		    }
		    else
		    {
			ResultsDataset predPruneData = new ResultsDataset();
			predPruneData.CreateDataset(svm,Pruning,path+"prune"+id+".pred");
			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(svm,Test,path+"testB"+id+".pred");
		    }
		    id++;
		}
	    }

            // NB 2 models           
            for (int nbCount=0; nbCount<2; nbCount++)
	    {
		NaiveBayes nb = new NaiveBayes();
		if (nbCount == 1)
		    nb.setUseKernelEstimator(true);


		try{
		    nb.buildClassifier(data);
		}
		    catch(Exception ex){ex.printStackTrace();System.out.println(ex);}
		if(type)
		{
		    ResultsDataset predTrainData = new ResultsDataset();
		    predTrainData.CreateDataset(nb,Train80,path+"train"+id+".pred");
		    ResultsDataset predTestData = new ResultsDataset();
		    predTestData.CreateDataset(nb,Test,path+"test"+id+".pred");
		}
		else
		{
		    ResultsDataset predPruneData = new ResultsDataset();
		    predPruneData.CreateDataset(nb,Pruning,path+"prune"+id+".pred");
		    ResultsDataset predTestData = new ResultsDataset();
		    predTestData.CreateDataset(nb,Test,path+"testB"+id+".pred");
		}
		id++;
	    }
            
            // J48 2x2=4 models
            double[] cfactor = {0.25, 0.50};
            for(int lapCounter=0;lapCounter<2;lapCounter++)
	    {
		for(int cfCounter=0; cfCounter<cfactor.length; cfCounter++)
		{
		    J48 j48Model = new J48();
		    if(lapCounter == 0)
			j48Model.setUseLaplace(true);
		    else
			j48Model.setUseLaplace(false);

		    j48Model.setConfidenceFactor((float) cfactor[cfCounter]);

		try{
		    j48Model.buildClassifier(data);
		}
		    catch(Exception ex){ex.printStackTrace();System.out.println(ex);}
		    if(type)
		    {
			ResultsDataset predTrainData = new ResultsDataset();
			predTrainData.CreateDataset(j48Model,Train80,path+"train"+id+".pred");
			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(j48Model,Test,path+"test"+id+".pred");
		    }
		    else
		    {
			ResultsDataset predPruneData = new ResultsDataset();
			predPruneData.CreateDataset(j48Model,Pruning,path+"prune"+id+".pred");
			ResultsDataset predTestData = new ResultsDataset();
			predTestData.CreateDataset(j48Model,Test,path+"testB"+id+".pred");
		    }

		    id++;
		}
	    }


    }

    public static void main(String[] args)
    {
	ProduceModels pmodels = new ProduceModels();
	String source = null;
	String target = null;
	String file = null;
	Random rand = new Random(System.currentTimeMillis());

	pmodels.TR = Double.parseDouble(args[0]);
	pmodels.PR = Double.parseDouble(args[1]);
	pmodels.TE = Double.parseDouble(args[2]);
	pmodels.M = Integer.parseInt(args[3]);
	file = args[4];
	source = args[5];
	target = args[6];

	System.out.println("Building the Models for "+file+" ...");
	for(int i=1;i<=1;i++)
	{
	    pmodels.CreateData(file,source,target+"exp"+i+"/",rand);
	    System.out.println("Fold :"+i+" finished");
	}
//	double t=0.88578989098;
//	System.out.println(Math.round(t*1000));
    }
}
