/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: </p>
 * @author Grigorios Tsoumakas
 * @version 20/12/2006
 */

import weka.core.*;
import weka.classifiers.*;

import java.io.*;

public class ResultsDataset
{
    int numModels;
    int m_NumClasses;
    // actual x predicted
    double [][] m_ConfusionMatrix;
    Instances Data;
    int [] numTrue;
    int [] numFalse;

    public int classValue(int indexOfInstance)
    {
	return (int) Data.instance(indexOfInstance).classValue();
    }

    public int classifyInstance(int indexOfInstance)
    {
	double[] preds2 = new double[Data.numAttributes()];
	double[] preds = new double[Data.numAttributes()-1];
	preds2 = Data.instance(indexOfInstance).toDoubleArray();
	for (int i=0; i<preds.length; i++)
	    preds[i] = preds2[i];

	RandomIndexOfMax rnd = new RandomIndexOfMax();
	return rnd.getIndex(preds);
    }

    public void CreateDataset(Classifier aClassifier, Instances testData, String Filename)
    {
	// Prepare Predictions Instances
	int numClasses = testData.numClasses();

	FastVector attInfo = new FastVector(0);
	for (int i=0; i<numClasses; i++)
	{
	    Attribute Att = new Attribute("class" + i);
	    attInfo.addElement(Att);
	}
	Attribute ClassAtt = testData.classAttribute();
	attInfo.addElement(ClassAtt);
	Instances predData = new Instances("Predictions", attInfo, 0); // add the classifier description

	// Add the predictions and the true class at the end
	double[] pred = new double[numClasses];
	int numInstances = testData.numInstances();
	for (int i=0; i<numInstances; i++)
	{
	    try {
		pred = aClassifier.distributionForInstance(testData.instance(i));
	    } catch (Exception ex) {
		ex.printStackTrace();
	    }
	    double[] pred2 = new double[numClasses+1];
	    for (int j=0; j<numClasses; j++)
		pred2[j] = pred[j];
	    pred2[numClasses] = testData.instance(i).classValue();          
	    predData.add(new Instance(1, pred2));
	}
	// Write Predictions
	try {
	    BufferedWriter FilePredictions = new BufferedWriter(new FileWriter(Filename));
	    FilePredictions.write(predData.toString());
	    FilePredictions.close();
	} catch (IOException ex) {
	    ex.printStackTrace();
	}
    }


    public void ZeroPredictions()
    {
	int num_ins = Data.numInstances();
	int num_attr = Data.numAttributes()-1;

	for (int i=0; i<num_ins; i++)
	{
	    numTrue[i] = 0;
	    numFalse[i] = 0;
	    for (int j=0; j<num_attr; j++)
		Data.instance(i).setValue(j, 0);
	}

	numModels=0;
    }

    public void AddPredictions(Instances Predictions, double Weight)
    {
	for (int i=0; i<Data.numInstances(); i++)
	    for (int j=0; j<Data.numAttributes()-1; j++)
		Data.instance(i).setValue(j, Data.instance(i).value(j)+Predictions.instance(i).value(j)*Weight);
    }


    public void AddPredictions(Instances Predictions)
    {
	numModels+=1;

	ResultsDataset tempr = new ResultsDataset(Predictions);

	int num_ins = Data.numInstances();
	int num_attr = Data.numAttributes()-1;

	for (int i=0; i<num_ins; i++)
	{
	    if(tempr.classifyInstance(i)==tempr.classValue(i))
		numTrue[i]+=1;
	    else
		numFalse[i]+=1;

	    for (int j=0; j<num_attr; j++)
		Data.instance(i).setValue(j, Data.instance(i).value(j)+Predictions.instance(i).value(j));
	}
    }

    public void RemovePredictions(Instances Predictions)
    {
	numModels-=1;

	ResultsDataset tempr = new ResultsDataset(Predictions);
	
	for (int i=0; i<Data.numInstances(); i++)
	{
	    if(tempr.classifyInstance(i)==tempr.classValue(i))
		numTrue[i]-=1;
	    else
		numFalse[i]-=1;

	    for (int j=0; j<Data.numAttributes()-1; j++)
		Data.instance(i).setValue(j, Data.instance(i).value(j)-Predictions.instance(i).value(j));
	}
    }

    public void DividePredictions(int Num)
    {
	if (Num == 1)
	    return;

	for (int i=0; i<Data.numInstances(); i++)
	    for (int j=0; j<Data.numAttributes()-1; j++)
		Data.instance(i).setValue(j, Data.instance(i).value(j)/Num);
    }

    public ResultsDataset()
    {
	numModels = 0;
    }

    public ResultsDataset(Instances Predictions, Instances Test)
    {
	Test.setClassIndex(Test.numAttributes()-1);
	Data = new Instances(Predictions);
	Data.insertAttributeAt(Test.attribute(Test.numAttributes()-1),
		Data.numAttributes());
	Data.setClassIndex(Data.numAttributes()-1);
	for (int i=0; i<Test.numInstances(); i++)
	    Data.instance(i).setClassValue(Test.instance(i).classValue());
	//      System.out.println(Data.toString());
	m_NumClasses = Data.numClasses();
	m_ConfusionMatrix = new double[m_NumClasses][m_NumClasses];
	calcConfusionMatrix();
    }

    public ResultsDataset(ResultsDataset x)
    {
	numModels = x.numModels;
	Data = new Instances(x.Data);
	int size = x.numTrue.length;
	numTrue = new int[size];
	numFalse = new int[size];
	m_NumClasses = x.Data.numClasses();
	m_ConfusionMatrix = new double[m_NumClasses][m_NumClasses];
	for(int i=0;i<size;i++)
	{
	    numTrue[i] = x.numTrue[i];
	    numFalse[i] = x.numFalse[i];
	}
    }

    public ResultsDataset(Instances Data)
    {
	numModels = 1;
	this.Data = new Instances(Data);
	int num_ins = Data.numInstances();
	m_NumClasses = Data.numClasses();
	m_ConfusionMatrix = new double[m_NumClasses][m_NumClasses];
	numTrue = new int[num_ins];
	numFalse = new int[num_ins];

	for(int i=0;i<num_ins;i++)
	{
	    if(this.classifyInstance(i)==this.classValue(i))
		numTrue[i]=1;
	    else
		numFalse[i]=1;
	}
	calcConfusionMatrix();
    }

    public void calcConfusionMatrix()
    {
	for (int i=0; i<m_NumClasses; i++)
	    for (int j=0; j<m_NumClasses; j++)
		m_ConfusionMatrix[i][j] = 0;

	for (int i=0; i<Data.numInstances(); i++)
	{
	    double PredictedDistribution[] = new double[m_NumClasses];
	    for (int j=0; j<m_NumClasses; j++)
		PredictedDistribution[j] = Data.instance(i).value(j);
	    int PredictedClass = Utils.maxIndex(PredictedDistribution);
	    int ActualClass = (int) Data.instance(i).classValue();

	    m_ConfusionMatrix[ActualClass][PredictedClass]++;
	}
    }

    /**
     * Calculate the number of true positives with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified positives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double numTruePositives(int classIndex) {

	double correct = 0;
	correct += m_ConfusionMatrix[classIndex][classIndex];
	return correct;
    }

    /**
     * Calculate the true positive rate with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified positives
     * ------------------------------
     *       total positives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double truePositiveRate(int classIndex) {

	double correct = 0, total = 0;
	for (int j = 0; j < m_NumClasses; j++) {
	    if (j == classIndex) {
		correct += m_ConfusionMatrix[classIndex][j];
	    }
	    total += m_ConfusionMatrix[classIndex][j];
	}
	if (total == 0) {
	    return 0;
	}
	return correct / total;
    }

    /**
     * Calculate the number of true negatives with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified negatives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double numTrueNegatives(int classIndex) {

	double correct = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i != classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j != classIndex) {
			correct += m_ConfusionMatrix[i][j];
		    }
		}
	    }
	}
	return correct;
    }

    /**
     * Calculate the true negative rate with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified negatives
     * ------------------------------
     *       total negatives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double trueNegativeRate(int classIndex) {

	double correct = 0, total = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i != classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j != classIndex) {
			correct += m_ConfusionMatrix[i][j];
		    }
		    total += m_ConfusionMatrix[i][j];
		}
	    }
	}
	if (total == 0) {
	    return 0;
	}
	return correct / total;
    }

    /**
     * Calculate number of false positives with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * incorrectly classified negatives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the false positive rate
     */
    public double numFalsePositives(int classIndex) {

	double incorrect = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i != classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j == classIndex) {
			incorrect += m_ConfusionMatrix[i][j];
		    }
		}
	    }
	}
	return incorrect;
    }

    /**
     * Calculate the false positive rate with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * incorrectly classified negatives
     * --------------------------------
     *        total negatives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the false positive rate
     */
    public double falsePositiveRate(int classIndex) {

	double incorrect = 0, total = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i != classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j == classIndex) {
			incorrect += m_ConfusionMatrix[i][j];
		    }
		    total += m_ConfusionMatrix[i][j];
		}
	    }
	}
	if (total == 0) {
	    return 0;
	}
	return incorrect / total;
    }

    /**
     * Calculate number of false negatives with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * incorrectly classified positives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the false positive rate
     */
    public double numFalseNegatives(int classIndex) {

	double incorrect = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i == classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j != classIndex) {
			incorrect += m_ConfusionMatrix[i][j];
		    }
		}
	    }
	}
	return incorrect;
    }

    /**
     * Calculate the false negative rate with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * incorrectly classified positives
     * --------------------------------
     *        total positives
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the false positive rate
     */
    public double falseNegativeRate(int classIndex) {

	double incorrect = 0, total = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i == classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    if (j != classIndex) {
			incorrect += m_ConfusionMatrix[i][j];
		    }
		    total += m_ConfusionMatrix[i][j];
		}
	    }
	}
	if (total == 0) {
	    return 0;
	}
	return incorrect / total;
    }

    /**
     * Calculate the recall with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified positives
     * ------------------------------
     *       total positives
     * </pre><p>
     * (Which is also the same as the truePositiveRate.)
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the recall
     */
    public double recall(int classIndex) {

	return truePositiveRate(classIndex);
    }

    /**
     * Calculate the overall accuracy
     *
     * @return the accuracy
     */

    public double accuracy() {
	calcConfusionMatrix();  
	double correct = 0;
	for (int i = 0; i < m_NumClasses; i++)
	    correct += numTruePositives(i);

	//      return correct / (positive(0)+negative(0));
	return correct / Data.numInstances();
    }


    /**
     * Calculate the precision with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * correctly classified positives
     * ------------------------------
     *  total predicted as positive
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the precision
     */
    public double precision(int classIndex) {

	double correct = 0, total = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i == classIndex) {
		correct += m_ConfusionMatrix[i][classIndex];
	    }
	    total += m_ConfusionMatrix[i][classIndex];
	}
	if (total == 0) {
	    return 0;
	}
	return correct / total;
    }

    /**
     * Calculate the F-Measure with respect to a particular class.
     * This is defined as<p>
     * <pre>
     * 2 * recall * precision
     * ----------------------
     *   recall + precision
     * </pre>
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the F-Measure
     */
    public double fMeasure(int classIndex) {

	double precision = precision(classIndex);
	double recall = recall(classIndex);
	if ((precision + recall) == 0) {
	    return 0;
	}
	return 2 * precision * recall / (precision + recall);
    }



    /**
     * Calculate the positives with respect to a particular class.
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double positive(int classIndex) {

	double total = 0;
	for (int j = 0; j < m_NumClasses; j++) {
	    total += m_ConfusionMatrix[classIndex][j];
	}
	return total;
    }


    /**
     * Calculate the negatives with respect to a particular class.
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the true positive rate
     */
    public double negative(int classIndex) {

	double total = 0;
	for (int i = 0; i < m_NumClasses; i++) {
	    if (i != classIndex) {
		for (int j = 0; j < m_NumClasses; j++) {
		    total += m_ConfusionMatrix[i][j];
		}
	    }
	}
	return total;
    }


    /**
     * Calculate the overall AUC
     *
     * @param Data is the Instances with the predictions
     * @return the AUC
     */
    public double getAUC(Instances Data, int MultiClassMethod) {
	if (m_NumClasses == 2)
	    return auc(Data, 0);
	else
	{
	    double auc = 0;

	    switch (MultiClassMethod) {
		case 0 : for (int i=0; i<m_NumClasses; i++) {
			     //System.out.println("AUC of reference class " + i + " = " + auc(Data, i));
			     if (auc(Data, i) > 0)
				 auc += auc(Data, i)*(positive(i) / (positive(i) + negative(i)));
			 }
			 break;
		case 1 : for (int i=0; i<m_NumClasses-1; i++)
			     for (int j=i+1; j<m_NumClasses; j++)
				 if (auc(Data, i, j) > 0)
				     auc += auc(Data, i, j);
			 auc = (2*auc)/((m_NumClasses-1)*m_NumClasses);
			 break;
	    }

	    return auc;
	}
    }

    /**
     * Calculate the AUC with respect to two classes.
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the AUC
     */
    public double auc(Instances Data, int class1, int class2) {
	Instances TempData = new Instances(Data, 0);

	for (int i=0; i<Data.numInstances(); i++)
	    if ((int) Data.instance(i).classValue() == class1 || (int) Data.instance(i).classValue() == class2)
		TempData.add(Data.instance(i));

	return auc(TempData, 0);
    }

    /**
     * Calculate the AUC with respect to a particular class.
     *
     * @param classIndex the index of the class to consider as "positive"
     * @return the AUC
     */
    public double auc(Instances Data, int classIndex) {
	Instances SortedData = new Instances(Data, 0);

	Data.sort(classIndex);
	for (int i=Data.numInstances()-1; i>=0; i--)
	    SortedData.add(Data.instance(i));

	double FP, TP, FP_Prev, TP_Prev, A, f_Prev, f_i, P, N;

	P = positive(classIndex);
	N = negative(classIndex);
	FP=0; TP=0;
	FP_Prev=0; TP_Prev=0;
	A=0;
	f_Prev=-1;

	for (int i=0; i<SortedData.numInstances(); i++) {
	    f_i = SortedData.instance(i).value(classIndex);
	    if (f_i != f_Prev) {
		A += TrapezoidArea(FP, FP_Prev, TP, TP_Prev);
		f_Prev = f_i;
		FP_Prev = FP;
		TP_Prev = TP;
	    }
	    if ((int) SortedData.instance(i).classValue() == classIndex)
		TP++;
	    else
		FP++;
	}
	A += TrapezoidArea(N, FP_Prev, P, TP_Prev);

	if (P*N > 0)
	    A = A/(P*N);
	else
	    A = -1;

	return A;
    }

    private double TrapezoidArea(double X1, double X2, double Y1, double Y2) {
	double Base = Math.abs(X1-X2);
	double HeightAvg = (Y1+Y2)/2.0;

	return (Base*HeightAvg);
    }

}
